/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.panels.request.components.editor;

import java.awt.BorderLayout;
import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.eviware.soapui.support.components.Inspector;
import com.eviware.soapui.support.components.JInspectorPanel;
import com.eviware.soapui.support.components.VTextIcon;
import com.eviware.soapui.support.components.VerticalTabbedPaneUI;

/**
 * Editor-framework for Xml Documents
 * 
 * @author ole.matzura
 */

@SuppressWarnings("serial")
public class XmlEditor extends JPanel implements PropertyChangeListener, XmlLocationListener
{
	private JTabbedPane inputTabs;
	private List<XmlEditorView> views = new ArrayList<XmlEditorView>();
	private XmlEditorView currentView;
	private XmlDocument xmlDocument;
	private JInspectorPanel inspectorPanel;
	private InputTabsChangeListener inputTabsChangeListener;

	public XmlEditor( XmlDocument xmlDocument )
	{
		super( new BorderLayout() );
		this.xmlDocument = xmlDocument;
		
		setBackground(Color.LIGHT_GRAY);
	   inputTabs = new JTabbedPane(JTabbedPane.LEFT, JTabbedPane.SCROLL_TAB_LAYOUT);
   	inputTabs.setUI(new VerticalTabbedPaneUI());
	   
		inputTabs.setFont(inputTabs.getFont().deriveFont(8));
		inputTabsChangeListener = new InputTabsChangeListener();
		inputTabs.addChangeListener( inputTabsChangeListener );
		
		inspectorPanel = new JInspectorPanel( inputTabs );
		add( inspectorPanel, BorderLayout.CENTER );
	}
	
	public void addEditorView( XmlEditorView editorView )
	{
		views.add( editorView );
		
		inputTabs.addTab( null, new VTextIcon(inputTabs, editorView.getTitle(), VTextIcon.ROTATE_LEFT ),
				editorView.getComponent() );

		editorView.addPropertyChangeListener( this );
		editorView.addLocationListener( this );
		
		editorView.setXmlDocument( xmlDocument );
	}

	public void propertyChange(PropertyChangeEvent evt)
	{
		if( evt.getPropertyName().equals( XmlEditorView.TITLE_PROPERTY ))
		{
			int ix = views.indexOf( evt.getSource() );
			if( ix == -1 )
				return;
			
			inputTabs.setTitleAt( ix, (String) evt.getNewValue() );
		}
	}
	
	public void selectView( int viewIndex )
	{
		inputTabs.setSelectedIndex( viewIndex );
	}
	
	public void selectView( String viewId )
	{
		for( int c = 0; c < views.size(); c++ )
		{
			if( views.get( c ).getViewId().equals( viewId ))
			{
				inputTabs.setSelectedIndex( c );
				return;
			}
		}
	}
	
	
	public void requestFocus()
	{
		if( currentView != null )
			currentView.getComponent().requestFocus();
	}

	public final XmlDocument getXmlDocument()
	{
		return xmlDocument;
	}
	
	public boolean saveDocument( boolean validate )
	{
	   return currentView == null ? true : currentView.saveDocument( validate );
	}
	
	public boolean hasFocus()
	{
		return currentView == null ? false : currentView.getComponent().hasFocus();
	}

	public final void setXmlDocument(XmlDocument xmlDocument)
	{
		if( this.xmlDocument != null )
			this.xmlDocument.release();
		
		this.xmlDocument = xmlDocument;
		
		for( XmlEditorView view : views )
		{
			view.setXmlDocument( xmlDocument );
		}
	}

	public final XmlEditorView getCurrentView()
	{
		return currentView;
	}

	public final JTabbedPane getInputTabs()
	{
		return inputTabs;
	}

	public final List<XmlEditorView> getViews()
	{
		return views;
	}
	
	public XmlEditorView getView( String viewId )
	{
		for( XmlEditorView view : views )
		{
			if( view.getViewId().equals( viewId ))
				return view;
		}
		
		return null;
	}
	
	public Inspector getInspector( String inspectorId  )
	{
		return inspectorPanel.getInspector( inspectorId );
	}

	public void locationChanged(XmlLocation location)
	{
		if( location != null )
		{
			for( Inspector inspector : inspectorPanel.getInspectors() )
			{
				((XmlInspector)inspector).locationChanged( location );
			}
		}
	}
	
	public void setEditable( boolean enabled )
	{
		for( XmlEditorView view : views )
		{
			view.setEditable( enabled );
		}
	}
	
	public void addInspector( XmlInspector inspector )
	{
		inspectorPanel.addInspector( inspector );
	}
	
	private final class InputTabsChangeListener implements ChangeListener
	{
		private int lastDividerLocation;

		public void stateChanged( ChangeEvent e )
		{
			int currentViewIndex = views.indexOf( currentView );
			
			if( currentView != null )
			{
				if( inputTabs.getSelectedIndex() == currentViewIndex)
					return;
				
				if( !currentView.deactivate() )
				{
					inputTabs.setSelectedIndex( currentViewIndex );
					return;
				}
			}
			
			XmlEditorView previousView = currentView;
			int selectedIndex = inputTabs.getSelectedIndex();
			if( selectedIndex == -1 )
			{
				currentView = null;
				return;
			}
			
			currentView = views.get( selectedIndex );
			
			if( currentView != null && !currentView.activate( previousView == null ? null : previousView.getLocation() ))
			{
				inputTabs.setSelectedIndex( currentViewIndex );
				if( currentViewIndex == -1 )
					return;
			}
			
			XmlInspector currentInspector = ( XmlInspector ) inspectorPanel.getCurrentInspector();
			
			if( !currentView.isInspectable() && currentInspector != null )
				lastDividerLocation = inspectorPanel.getDividerLocation();
			
			inspectorPanel.setInspectorVisible( currentView.isInspectable() && currentInspector != null );
			inspectorPanel.setToolbarVisible( currentView.isInspectable()  );
			
			if( currentView.isInspectable() && currentInspector != null )
			{
				if( lastDividerLocation == 0 )
					inspectorPanel.setResetDividerLocation();
				else
					inspectorPanel.setDividerLocation( lastDividerLocation );
			}
			
			SwingUtilities.invokeLater( new Runnable() {
		
				public void run()
				{
					if( currentView != null )
						currentView.getComponent().requestFocus();
				}} );
		}
	}
	
	public void release()
	{
		for( XmlEditorView view : views )
		{
			view.release();
			view.removeLocationListener( this );
			view.removePropertyChangeListener( this );
		}
		
		views.clear();
		
		inputTabs.removeChangeListener( inputTabsChangeListener );
		inputTabs.removeAll();
		
		inspectorPanel.release();
		xmlDocument.release();
	}
}
